package de.zkm.pdf.process;

import java.util.ArrayList;
import java.util.List;

import de.zkm.pdf.model.PageTextBoxData;
import de.zkm.pdf.model.TextBoxData;
import de.zkm.pdf.model.TextBoxType;
import lombok.extern.slf4j.Slf4j;

@Slf4j
public class PageTextBoxDataProcessor {

	/**
	 * Concatenates all {@link PageTextBoxData#textBoxes} that belong to the same line into
	 * one {@link TextBoxData} per line.
	 * <p>
	 * Expects the {@link PageTextBoxData#textBoxes} to ordered by {@link TextBoxData#lineIndex}
	 * @param pageTextBoxData
	 * @return
	 */
	public PageTextBoxData joinWordsToLines(PageTextBoxData pageTextBoxData) {
		
		PageTextBoxData linePageData = pageTextBoxData.clone();
		int currentLineindex = -1;
		List<TextBoxData> lineBoxDataList = new ArrayList<>();
		for(TextBoxData boxdata : pageTextBoxData.getTextBoxes()) {
			if(currentLineindex != boxdata.getLineIndex()) {
				// new line detected
				if(currentLineindex > 0) { // lineIndex is 1 based!, see PageTextBoxData
					// not before end of first line
					try {
						textBoxData2Line(linePageData, lineBoxDataList);
					} catch (PageTextBoxProcessorException e) {
						log.error("Error processing last line of " + pageTextBoxData.getUri() + " at line " + currentLineindex, e);
					}
					lineBoxDataList = new ArrayList<>();
				}
				currentLineindex = boxdata.getLineIndex();
			}
			lineBoxDataList.add(boxdata);
		}
		
		// Also add the last line
		try {
			textBoxData2Line(linePageData, lineBoxDataList);
		} catch (PageTextBoxProcessorException e) {
			log.error("Error processing last line of " + pageTextBoxData.getUri() + " at line " + currentLineindex, e);
		}
		
		return linePageData;

	}

	protected void textBoxData2Line(PageTextBoxData linePageData, List<TextBoxData> lineBoxDataList) throws PageTextBoxProcessorException {
			// remove trailing whitespace
			while(lineBoxDataList.get(lineBoxDataList.size() - 1).getText().matches("\\s+")) {
				lineBoxDataList.remove(lineBoxDataList.size() - 1);
			}
			linePageData.addTextBox(concat(lineBoxDataList));
	}

	private TextBoxData concat(List<TextBoxData> lineBoxDataList) throws PageTextBoxProcessorException {
		TextBoxData lineData = new TextBoxData();
		lineData.setType(TextBoxType.line);
		if(lineBoxDataList.isEmpty()) {
			return lineData;
		}
		TextBoxData firstWord = lineBoxDataList.get(0);
		lineData.setLineIndex(firstWord.getLineIndex());
		lineData.setLeft(firstWord.getLeft());
		lineData.setTop(firstWord.getTop());
		for(TextBoxData wordData : lineBoxDataList) {
			// validity check
			if(lineData.getLineIndex() != wordData.getLineIndex()) {
				throw new PageTextBoxProcessorException("Alternating line index in lineBoxDataList: " 
						+ lineData.getLineIndex() + " -> " + wordData.getLineIndex());
			}
			lineData.getTextBuilder().append(wordData.getText());
			lineData.setHeight(Math.max(lineData.getHeight(), wordData.getHeight()));
			lineData.setWidth(lineData.getWidth() + wordData.getHeight());
		}
		return lineData;
	}
}
