package de.zkm.pdf.text;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.text.PDFTextStripper;
import org.apache.pdfbox.text.TextPosition;

import de.zkm.pdf.model.PageTextBoxData;
import de.zkm.pdf.model.TextBoxData;
import de.zkm.pdf.model.TextBoxType;
import de.zkm.pdf.model.WordBounds;
import lombok.extern.slf4j.Slf4j;

public class PositionalPDFTextStripper extends PDFTextStripper {

	float pageUserUnit = 0f;
	
	private List<PageTextBoxData> pageDataList = new ArrayList<>();
	private URI pdfFileURI;
	/**
	 * 1-based line index since the page indices also start at 1 
	 */
	private int START_LINE_INDEX = 1;
	private int lineIndex = START_LINE_INDEX;
	
	private PageTextBoxData pageData = null;

	public PositionalPDFTextStripper() {
		super();
		setSortByPosition(true);
	}
    
    @Override
    public void processPage(PDPage page) throws IOException {
    	if(pageData != null){
    		// add data for last page to list
    		pageDataList.add(pageData);
    	}
    	pageData = new PageTextBoxData(pdfFileURI, getEndPage());
    	pageUserUnit = page.getUserUnit();
    	pageData.setPageIndex(getCurrentPageNo());
    	pageData.setWidth(scale(page.getCropBox().getWidth()));
    	pageData.setHeight(scale(page.getCropBox().getHeight()));
    	// reset the line index
    	lineIndex = START_LINE_INDEX;
    	// call super method at last
    	super.processPage(page);
    }
    
    protected void endDocument(PDDocument document) throws IOException  {
    	// add data for last page processed
    	pageDataList.add(pageData);
    }
    
    protected float scale(float value) {
    	return value * pageUserUnit;
    }
    
    protected WordBounds calculateBounds(List<TextPosition> textPositions) {
    	WordBounds wb  = new WordBounds();
    	if(textPositions.size() > 0) {
    		wb.x = textPositions.get(0).getXDirAdj();
    		wb.y = textPositions.get(0).getYDirAdj();
    		for(TextPosition pos : textPositions) {
    			wb.w = wb.w + pos.getWidthDirAdj();
    			wb.h = Math.max(wb.h, pos.getHeightDir());
    			wb.word.append(pos.getUnicode());
    		}
    	}
    	wb.x = scale(wb.x);
    	wb.y = scale(wb.y);
    	wb.w = scale(wb.w);
    	wb.h = scale(wb.h);
    	return wb;
    }
    
    /**
     * Detects the TextBoxType of the TextPosition. Only returns {@link TextBoxType#word},
     * {@link TextBoxType#punctuation}, or {@link TextBoxType#punctuation}, but <b>never</b>
     * {@link TextBoxType#phrase}, as phrases are constructed in a post peocessing step from 
     * all TextPositions of different types that together make up a line. 
     * 
     * @param position
     * @return
     */
    protected TextBoxType typeOf(TextPosition position) {
    	if(position.getUnicode().equals(getWordSeparator())){
    		return TextBoxType.word_separator;
    	}
    	if(position.getUnicode().equals(getLineSeparator())){
    		return TextBoxType.line_separator;
    	}
    	if(position.getUnicode().matches("\\p{Punct}")) {
    		return TextBoxType.punctuation;
    	}
    	return TextBoxType.word;
    }
    
    public List<PageTextBoxData> getPageData(URI pdfFileURI, PDDocument document) throws IOException {
    	pageDataList = new ArrayList<>();
		this.pdfFileURI = pdfFileURI;
    	setEndPage(document.getNumberOfPages());
    	getText(document);
    	return pageDataList;
    }
    
    protected void writeLineSeparator() throws IOException
    {
    	// increment line index number
    	lineIndex++;
    }
    
    @Override
    protected void writeString(String text, List<TextPosition> textPositions) throws IOException {
    	
    	List<TextPosition> wordPositions = new ArrayList<TextPosition>();
    	
    	for (TextPosition position : textPositions)
		{
    		TextBoxType currentType = typeOf(position);
    		if(currentType.equals(TextBoxType.word)) {
    			wordPositions.add(position);
    		} else {
    			// compose and add word
    			if(!wordPositions.isEmpty()) {
	    			WordBounds wb = calculateBounds(wordPositions);
			    	TextBoxData textBoxData = new TextBoxData(); 
			    	textBoxData.setLineIndex(lineIndex);
			    	textBoxData.setType(TextBoxType.word);
			    	textBoxData.setLeft(Math.round(wb.x));
			    	textBoxData.setTop(Math.round(wb.y));
			    	textBoxData.setWidth(Math.round(wb.w));
			    	textBoxData.setHeight(getGraphicsStackSize());
			    	textBoxData.setText(wb.word.toString());
			    	pageData.addTextBox(textBoxData);
			    	wordPositions.clear();
    			}
		    	// " fs=" + position.getFontSize() +
		    	// " space=" + scale(position.getWidthOfSpace()) + 
		    	// " xscale=" + position.getXScale() +
		    	// add punctuation or word separator
		    	TextBoxData punctOrSep = new TextBoxData(); 
		    	punctOrSep.setType(currentType);
		    	punctOrSep.setLineIndex(lineIndex);
		    	punctOrSep.setLeft(scale(position.getXDirAdj()));
		    	punctOrSep.setTop(scale(position.getYDirAdj()));
		    	punctOrSep.setWidth(scale(position.getWidthDirAdj()));
		    	punctOrSep.setHeight(scale(position.getHeightDir()));
		    	punctOrSep.setText(position.getUnicode());		    	
		    	pageData.addTextBox(punctOrSep);
    		}
	    
		}
    }
}